/*
 * MPEG-1/2 encoder header
 * Copyright (c) 2007 Aurelien Jacobs <aurel@gnuage.org>
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifndef AVCODEC_MPEG12ENC_H
#define AVCODEC_MPEG12ENC_H

#include "mpegvideoenc.h"

void ff_mpeg1_encode_slice_header(MPVEncContext *s);

static inline void ff_mpeg1_clean_buffers(MPVEncContext *s)
{
    s->last_dc[0] = 128 << s->c.intra_dc_precision;
    s->last_dc[1] = s->last_dc[0];
    s->last_dc[2] = s->last_dc[0];
    memset(s->c.last_mv, 0, sizeof(s->c.last_mv));
}

#endif /* AVCODEC_MPEG12ENC_H */
